# coding=utf8

"""
author : MaHaoYang
publish date : 2018-12-21
name : link rank
expression : Rank = ∑ (ln((ln(TextLength+1) + TagMatrix * WeightMatrix) * α)) ± δ;  α = 2 ; δ ≈ 5 * 10e-3

"""

import re
import copy
import math
import traceback

import numpy as np
import cchardet
import lxml
import lxml.html
from lxml import etree
from lxml.html import HtmlComment

from spider_tool.format import Format
from spider_tool.clean import Clean

'''
relevance_html_tags = \
    [
        'nav',
        'a',

        'title'
        'div',
        'span',

        'ul',
        'ol',
        'li',
        'dl',
        'dt',
        'dd',
        'menu',

        'table',
        'caption',
        'th',
        'tr',
        'td',
        'thread',
        'tbody',
        'tfoot',

        'b',
        'p',

    ]
'''


class NewsLinkExtractor:
    relevance_html_tags = \
        [
            'a',

            'div',
            'ul',
            'ol',
            'li',
            'dl',
            'dt',
            'dd',

            'table',
            'tbody',

            'tr',
            'td',

        ]
    # ['a', 'div', 'ul', 'ol', 'li', 'dl', 'dt', 'dd', 'table', 'tbody', 'tr', 'td', ]
    relevance_weight = [3, -1, 4, 4, 3, 4, 2, 2, 4, 4, 3, 3]
    find_zh = re.compile(u'[\u4e00-\u9fa5]').findall
    find_zh_en = re.compile(u'[a-zA-Z\d_\u4e00-\u9fa5]').findall
    layer_num = None
    tag_layer_num = None
    longtext_layer_num = None
    text_num = None
    date_num = None
    all_chs_num = None
    char_chain = None

    def _is_sub_string(self, content, parent=None):
        if parent is None:
            parent = self.char_chain
        for i in range(len(parent)):
            if parent[i] == content[0]:
                if content == parent[i:i + len(content)]:
                    return True
                else:
                    return False
        else:
            return False

    def _calc_tag(self, node, tags=None, lower_bound=7):
        if tags is None:
            tags = self.relevance_html_tags
        for child in node:
            self.layer_num += 1
            if child.tag in tags:
                self.tag_layer_num[self.relevance_html_tags.index(child.tag)] += 1
                child_text = etree.tostring(child, encoding='utf-8').decode()
                if self.find_zh(child_text):
                    child_text = ''.join(self.find_zh(child_text))
                    if not self._is_sub_string(child_text):
                        self.char_chain.append(child_text)
                    # fixme : 只统计中文
                    if (len(self.find_zh(child_text)) > lower_bound) and (child.tag not in ['p', 'div']):
                        self.longtext_layer_num += 1

                    # fixme : 只统计中文
                    self.text_num += len(self.find_zh(child_text))
                    # fixme : date extraction too slow
            try:
                date_flag = Format().time_normalization(child.text)
            except Exception as E:
                date_flag = 0
            if date_flag:
                self.date_num += 1
            if child.tag:
                self._calc_tag(child, tags=tags)

    def calc_tag(self, node, tags=None, lower_bound=7):
        # 各子循环之间的全局共享参数，所以每个循环之前需要初始化
        self.layer_num = 0.0
        self.tag_layer_num = [0.0] * len(self.relevance_html_tags)
        self.longtext_layer_num = 0.0
        self.text_num = 0.0
        self.date_num = 0.0
        self.char_chain = list()

        self._calc_tag(node, tags=tags, lower_bound=lower_bound)
        tag_density = [i / (self.layer_num + 1) for i in self.tag_layer_num]
        longtext_density = self.longtext_layer_num / (self.layer_num + 1)
        char_level_density = self.text_num / (self.layer_num + 1)
        date_density = self.date_num / (self.layer_num + 1)
        weight = dict()
        weight['td'] = tag_density
        weight['tn'] = self.tag_layer_num
        weight['ltd'] = longtext_density
        weight['ltn'] = self.longtext_layer_num
        weight['cld'] = char_level_density
        weight['cln'] = self.text_num
        weight['dd'] = date_density
        weight['dn'] = self.date_num
        chsn = len(self.find_zh(etree.tostring(node, encoding='utf-8').decode('utf-8')))
        weight['chsn'] = chsn if chsn > 20 else 0
        chsr = float(chsn) / float(self.all_chs_num)
        weight['chsr'] = 1.3 - chsr if chsr > 0.667 else chsr
        weight['chsd'] = chsn / len(etree.tostring(node, encoding='utf-8').decode('utf-8'))
        return weight

    candidate = None
    node = None
    tag_filter = ['ul', 'ol', 'dl', 'table', 'tbody', 'div']
    _layer_depth = 0
    layer_depth = None

    def _calc_tags(self, node):
        self._layer_depth += 1
        if self.candidate is None:
            self.candidate = []
            self.node = []
            self.layer_depth = []
        for child in node:
            if child.tag:
                if child.tag in self.tag_filter:
                    candi = self.calc_tag(child)
                    self.candidate.append(candi)
                    self.node.append(child)
                    self.layer_depth.append(self._layer_depth)
                self._calc_tags(child)

    def calc_tags(self, node):
        self._calc_tags(node)
        return self.candidate

    def calc_score(self, url, node):
        node = self.normalize_dom(url, node)
        self.calc_tags(node)
        res = []
        self.relevance_weight = np.array(self.relevance_weight)
        candidate = list(zip(self.candidate, self.layer_depth))
        for candi in candidate:
            # 逐个打分
            weight = self.relevance_weight.dot(np.array([math.log(i + 1) for i in candi[0]['tn']])) * (
                candi[0]['tn'][0] if candi[0]['tn'][0] > 4 else 0)
            # score = (weight if weight > 0 else math.exp(weight)) * math.exp(candi[0]['ltn']) * candi[0]['cln'] * (
            #         candi[0]['dn'] + 1)
            score = math.exp(math.log(max(weight, 1.001))) * candi[0]['ltd'] * candi[0]['cld'] * candi[0][
                'chsn'] * math.exp(candi[0]['chsr'] * candi[0]['chsd'] * 100) * math.exp(candi[0]['dd'])
            score = math.log(math.log(score + 1, 10e4) ** (math.log(candi[1] + 1, 2)) + 1, 10e4)
            res.append(score)
        # result = list(zip(res, [etree.tostring(i, encoding='utf-8').decode() for i in self.node]))
        result = list(zip(res, self.node))
        result.sort(key=lambda e: e[0], reverse=True)
        return result

    def calc_top_link(self, url, node):
        if isinstance(node, bytes):
            node = node.decode('utf-8')
        self.all_chs_num = len(self.find_zh(node))
        result = self.calc_score(url, node)
        if not result:
            print('\tno main block got !!!!!', url)
            return '', ''
        result = result[0:1]
        for i in result:
            self.get_longtext_link(i[1])
        link_with_name = copy.deepcopy(self.longtext_link)
        if url in link_with_name.keys():
            link_with_name.pop(url)
        link_with_name = dict(zip(link_with_name.values(), link_with_name.keys()))
        self.longtext_link = None
        link_with_html = '\n'.join([etree.tostring(i[1], encoding='utf-8').decode() for i in result])
        return link_with_name, link_with_html

    longtext_link = None
    search_name = ''

    def _search_name(self, node):
        self.search_name = Clean.clean_all_blank(Clean.clean_html_tag(node.text))
        if self.search_name:
            if not len(self.find_zh(self.search_name)):
                for attr in node.attrib.values():
                    if len(self.find_zh(attr)):
                        self.search_name = Clean.clean_all_blank(Clean.clean_html_tag(attr))
                        return
                else:
                    for child in node:
                        self._search_name(child)
        else:
            for attr in node.attrib.values():
                if len(self.find_zh(attr)):
                    self.search_name = Clean.clean_all_blank(Clean.clean_html_tag(attr))
            else:
                for child in node:
                    self._search_name(child)

    def get_longtext_link(self, node):
        if self.longtext_link is None:
            self.longtext_link = dict()
        for child in node:
            if child.tag == 'a':
                link = child.get('href')
                self.search_name = ''
                self._search_name(child)
                name = self.search_name if self.search_name else ''
                if name:
                    if len(self.find_zh(name)) > 7:
                        self.longtext_link[link] = name
            self.get_longtext_link(child)

    @staticmethod
    def normalize_dom(url, dom_bytes):
        if isinstance(dom_bytes, str):
            dom_bytes = dom_bytes.encode('utf-8')
        if isinstance(dom_bytes, bytes):
            encoding = cchardet.detect(dom_bytes)['encoding']
            dom_bytes = dom_bytes.decode(encoding, 'ignore')
            doc = lxml.html.fromstring(dom_bytes)
            doc.make_links_absolute(base_url=url)
            doc = doc.xpath('//body')
            return doc


REGEXES = {
    'okMaybeItsACandidateRe': re.compile(
        'and|article|artical|body|column|main|shadow', re.I),
    'positiveRe': re.compile(
        ('article|arti|body|content|entry|hentry|main|page|'
         'artical|zoom|arti|context|message|editor|'
         'pagination|post|txt|text|blog|story'), re.I),
    'negativeRe': re.compile(
        ('copyright|combx|comment|com-|contact|foot|footer|footnote|decl|copy|'
         'notice|'
         'masthead|media|meta|outbrain|promo|related|scroll|link|pagebottom|bottom|'
         'other|shoutbox|sidebar|sponsor|shopping|tags|tool|widget'), re.I),
}


class NewsExtractor:
    relevance_html_tags = ['p', 'span', 'br', 'div', 'table', 'tbody', 'tr', 'td',
                           'a', 'h1', 'h2', 'h3', 'h4', 'h5']
    # ['p', 'div', 'ul', 'ol', 'dl', 'table', 'tbody', 'tr', 'td', ]
    relevance_weight = [10, 3, 1, 0.1, 2, 2, 2, 2,
                        -10, -5, -2, -2, -1, -1]
    find_zh = re.compile(u'[\u4e00-\u9fa5]').findall
    find_zh_en = re.compile(u'[a-zA-Z\d_\u4e00-\u9fa5]').findall
    layer_num = None
    tag_layer_num = None
    longtext_layer_num = None
    text_num = None
    date_num = None
    all_chs_num = None
    char_chain = None

    def _is_sub_string(self, content, parent=None):
        if parent is None:
            parent = self.char_chain
        for i in range(len(parent)):
            if parent[i] == content[0]:
                if content == parent[i:i + len(content)]:
                    return True
                else:
                    return False
        else:
            return False

    def _calc_tag(self, node, tags=None, lower_bound=10):
        if tags is None:
            tags = self.relevance_html_tags
        for child in node:
            self.layer_num += 1
            if child.tag in tags:
                self.tag_layer_num[self.relevance_html_tags.index(child.tag)] += 1
                child_text = etree.tostring(child, encoding='utf-8').decode()
                if self.find_zh(child_text):
                    child_text = ''.join(self.find_zh(child_text))
                    if not self._is_sub_string(child_text):
                        self.char_chain.append(child_text)
                        # fixme : 只统计中文
                        if len(child_text) > lower_bound and child.tag != 'a':
                            self.longtext_layer_num += 1
                        self.text_num += len(child_text)
            try:
                #  先清洗，不然很容易得不出日期
                date_flag = Format().time_normalization(Clean.clean_html_tag(child.text))
            except Exception as E:
                date_flag = 0
            if date_flag:
                self.date_num += 1
            if child.tag:
                self._calc_tag(child, tags=tags)

    def calc_tag(self, node, tags=None, lower_bound=10):
        # 各子循环之间的全局共享参数，所以每个循环之前需要初始化
        self.layer_num = 0.0
        self.tag_layer_num = [0.0] * len(self.relevance_html_tags)
        self.longtext_layer_num = 0.0
        self.text_num = 0.0
        self.date_num = 0.0
        self.char_chain = list()

        self._calc_tag(node, tags=tags, lower_bound=lower_bound)
        tag_density = [i / (self.layer_num + 1) for i in self.tag_layer_num]
        longtext_density = self.longtext_layer_num / (self.layer_num + 1)
        char_level_density = self.text_num / (self.layer_num + 1)
        date_density = self.date_num / (self.layer_num + 1)
        weight = dict()
        weight['td'] = tag_density
        weight['tn'] = self.tag_layer_num
        weight['ltd'] = longtext_density
        weight['ltn'] = self.longtext_layer_num
        weight['cld'] = char_level_density
        weight['cln'] = self.text_num
        weight['dd'] = date_density
        weight['dn'] = self.date_num
        chsn = len(self.find_zh(etree.tostring(node, encoding='utf-8').decode('utf-8')))
        weight['chsn'] = chsn if chsn > 20 else 0
        chsr = float(chsn) / float(self.all_chs_num)
        weight['chsr'] = 1.3 - chsr if chsr > 0.667 else chsr
        weight['chsd'] = chsn / len(etree.tostring(node, encoding='utf-8').decode('utf-8'))

        return weight

    candidate = None
    node = None
    tag_filter = {'div'}
    _layer_depth = 0
    layer_depth = None

    def _calc_tags(self, node):
        self._layer_depth += 1
        if self.candidate is None:
            self.candidate = []
            self.node = []
            self.layer_depth = []
        for child in node:
            if child.tag:
                candi = self.calc_tag(child)
                if candi['td'][3] + candi['td'][4] > candi['td'][2] * 1.2:
                    # 这行代码加入后需要注意——每次初始化运行的地方都要需要初始化一次标签过滤器，作用是发现一次触发一次，再由初始化重置回来
                    self.tag_filter.update(['table', 'tbody', 'td', 'tr'])
                if child.tag in self.tag_filter:
                    self.candidate.append(candi)
                    self.node.append(child)
                    self.layer_depth.append(self._layer_depth)
                self._calc_tags(child)

    def calc_tags(self, node):
        self._calc_tags(node)
        return self.candidate

    def calc_score(self, url, node):
        # 这里需要初始化一次标签过滤器
        self.tag_filter = {'div'}
        self.calc_tags(node)
        res = []
        self.relevance_weight = np.array(self.relevance_weight)
        candidate = list(zip(self.candidate, self.layer_depth))
        for candi in candidate:
            # 逐个打分
            weight = self.relevance_weight.dot(np.array([math.log(i + 1) for i in candi[0]['td']]))
            score = math.exp(weight) * candi[0]['ltd'] * candi[0]['cld'] * candi[0]['chsn'] * math.exp(
                candi[0]['chsr'] * candi[0]['chsd'] * 100) / math.exp(candi[0]['dn'])
            score = math.log(math.log(score + 1, 10e4) ** (math.log(candi[1] + 1, 2)) + 1, 10e4)
            res.append(score)
        result = list(zip(res, enumerate([etree.tostring(i, encoding='utf-8').decode() for i in self.node])))
        # result = list(zip(res, self.node))
        result.sort(key=lambda e: e[0], reverse=True)
        return result

    '''
    最后精排序一次，按文本中文字数，以最高分为基准字数，选一定字数比例范围内，短的那个文本

    '''

    def _election(self, result):
        parent = ''.join(self.find_zh(result[0][1][1]))
        res = [result[0][1]]
        end = int(len(result) / 20) if len(result) > 40 else 2
        for i in result[:end]:
            ichs = ''.join(self.find_zh(i[1][1])) if self.find_zh(i[1][1]) else ' '
            if self._is_sub_string(ichs, parent) or self._is_sub_string(parent, ichs):
                if 0.999 > len(self.find_zh(i[1][1])) / len(self.find_zh(parent)) > 0.7:
                    if len(self.find_zh(i[1][1])) / len(self.find_zh(parent)) > 1:
                        res.append(i[1])
                    elif len(self.find_zh(i[1][1])) / len(self.find_zh(parent)) < 1:
                        res.insert(0, i[1])
            if len(res) > 1:
                break
        return res[0]

    # def calc_content(self, url, node):
    #     result = self.calc_score(url, node)
    #     if not result:
    #         print('\tno main block got !!!!!', url)
    #         return '', ''
    #     result = self._election(result)
    #     return result

    def easy_title(self, doc, title_area):
        title = ''
        title_el = doc.xpath('//title')
        if title_el:
            title = title_el[0].text_content().strip()
            # if self._is_sub_string(''.join(self.find_zh(_title)), ''.join(self.find_zh(title_area))):
            #     title = _title
        if len(title) < 7:
            tt = doc.xpath('//meta[@name="title"]')
            if tt:
                title = tt[0].get('content', '')
        if len(title) < 7:
            tt = doc.xpath('//*[contains(@id, "title") or contains(@class, "title")]')
            if not tt:
                tt = doc.xpath('//*[contains(@id, "font01") or contains(@class, "font01")]')
            if not tt:
                tt = doc.xpath('//*[contains(@id, "tit") or contains(@class, "tit")]')
            for t in tt:
                ti = t.text_content().strip()
                if ti in title and len(ti) * 2 > len(title):
                    title = ti
                    break
                if len(ti) > 20:
                    continue
                if len(ti) > len(title) or len(ti) > 7:
                    title = ti
        return title

    def shorten_title(self, title):
        spliters = [' - ', '–', '—', '-', '|', '::']
        for s in spliters:
            if s not in title:
                continue
            tts = title.split(s)
            if len(tts) < 2:
                continue
            title = tts[0]
            break
        return title

    longtext_link = None
    search_name = ''

    def calc_main_block(self, url, node):
        if isinstance(node, bytes):
            node = node.decode('utf-8')
        self.all_chs_num = len(self.find_zh(node))
        node = self.normalize_dom(url, node)
        node_body = node.xpath('//body')[0]
        # content = self.calc_content(url, node_body)
        result = self.calc_score(url, node_body)
        if not result:
            print('\tno main block got !!!!!', url)
            content = ('', '')
        else:
            content = self._election(result)
        if content[0] > 5:
            date_candi = self.node[content[0] - 5:content[0]]
            n = 5
            while n + 5 < content[0]:
                date_candi.extend(self.node[content[0] - n + 5:content[0] - n])
                n += 5
        elif content[0]:
            date_candi = self.node[:content[0]]
        else:
            date_candi = self.node[:1]
        title_area = etree.tostring(node_body, encoding='utf-8').decode('utf-8')
        # title_area = ''.join([i[1][1] for i in result[1:11]])
        title = self.easy_title(node, title_area)
        title = self.shorten_title(title)
        pub_time = ''
        for i in date_candi:
            i = etree.tostring(i, encoding='utf-8').decode('utf-8')
            pub_time = Format().time_normalization(Clean.clean_html_tag(i))
            if pub_time:
                break

        return title, pub_time, content[1]

    @staticmethod
    def normalize_dom(url, dom_bytes):
        if isinstance(dom_bytes, str):
            dom_bytes = dom_bytes.encode('utf-8')
        if isinstance(dom_bytes, bytes):
            encoding = cchardet.detect(dom_bytes)['encoding']
            dom_bytes = dom_bytes.decode(encoding, 'ignore')
            doc = lxml.html.fromstring(dom_bytes)
            doc.make_links_absolute(base_url=url)
            return doc


if __name__ == '__main__':
    import requests
    from fake_useragent import UserAgent

    ua = UserAgent()
    headers = dict()
    headers['User-Agent'] = ua.random
    m = NewsLinkExtractor()
    # FIXME: How to identify the title of article not in title tag, but in h1 tag 'http://www.biodiscover.com/talk/investment.html'
    url = 'http://www.biodiscover.com/news/financing'
    # url = 'http://www.hebfda.gov.cn/CL0417/'
    # url = 'http://www.eisai.com.cn/cn/news.php'
    # url = 'http://finance.sina.com.cn/roll/index.d.html?cid=56592'
    # url = 'http://www.cd120.com/htmlnewszhaomuxinxi/index.jhtml'
    # url = 'http://www.ada.gov.cn/4117559.html'
    # url = 'http://scjg.tj.gov.cn/xwzx/gs/index.html'
    html = requests.get(url=url, headers=headers)
    html.encoding = html.apparent_encoding
    html = html.text
    e = m.calc_top_link(url, html)
    # print(e[1])
    a = ['%s %s' % (i, k) for i, k in zip(e[0].keys(), e[0].values())]
    print('\n'.join(a))
    print(len(e[0].keys()))

    m = NewsExtractor()
    # FIXME: How to identify the title of article not in title tag, but in h1 tag 'http://www.biodiscover.com/talk/investment.html'
    url = 'http://www.biodiscover.com/news/financing/733045.html'
    # url = 'http://www.hebfda.gov.cn/CL0037/'
    # url = 'http://www.hebfda.gov.cn/CL0225/70885.html'
    # url = 'http://www.eisai.com.cn/cn/newsDetails.php?id=281'
    # url = 'https://finance.sina.com.cn/stock/s/2019-02-13/doc-ihrfqzka5321960.shtml'
    # url = 'http://www.cd120.com/htmlnewszhaomuxinxi/85333.jhtml'
    # url = 'http://www.ada.gov.cn/4148577/18006548.html'
    # url = 'http://www.ada.gov.cn/4148577/15417241.html'
    # url = 'http://www.ada.gov.cn/4148577/17748413.html'
    # url = 'http://scjg.tj.gov.cn/jixian/zwgz/gstg/ypjyxkxx/37129.html'
    # url = 'http://scjg.tj.gov.cn/xwzx/gs/spgg/ypylqxbjspgg/37161.html'
    # url = 'http://scjg.tj.gov.cn/binhaixinqu/zwgz/gstg/ypGSPrz/37111.html'
    url = 'http://www.biodiscover.com/news/financing/733000.html'
    url = 'http://www.biodiscover.com/news/financing/731962.html'
    html = requests.get(url=url, headers=headers)
    html.encoding = html.apparent_encoding
    html = html.text
    e = m.calc_main_block(url, html)
    print(len(e))
    print(e)
