from scrapy.selector import Selector
from scrapy.http import HtmlResponse
from scrapy.linkextractors.lxmlhtml import LxmlLinkExtractor
from . import *
from .toolkit import *
import requests
from .metrics import EXTRACT_COUNTER
from .auto_extractor import NewsLinkExtractor, NewsExtractor


class Extractor(object):

    @staticmethod
    def web_collector(url, html, extract_server):
        url = url
        if isinstance(html, str):
            html = html.encode('utf-8')
        html = {'html': html}
        extract_response = requests.post(extract_server, data=html,
                                         headers={'Content-Type': 'application/x-www-form-urlencoded',
                                                  'accept-language': 'en'}, verify=False, timeout=50)
        EXTRACT_COUNTER.inc()  # point
        items = dict()
        scrapy_response = HtmlResponse(url, encoding='utf-8', body=extract_response.content)
        jsr = json.loads(scrapy_response.text)
        tc = NewsExtractor()
        tc_result = tc.calc_main_block(url=url, node=html)
        if jsr.get('title'):
            items['title'] = jsr.get('title')
        else:
            items['title'] = tc_result[0]
        if tc_result[1]:
            items['pub_time'] = tc_result[1]
        else:
            items['pub_time'] = jsr.get('time')[0:10]
        if tc_result[2]:
            items['content_html'] = Clean.clean_blank_line(tc_result[2])
        else:
            items['content_html'] = Clean.clean_blank_line(jsr.get('html'))
        items['content'] = Clean.clean_html_tag(items['content_html'])
        return items

    @staticmethod
    def manual_rule(url, column_name, xpath, regex, html, is_detail, data):
        if isinstance(html, str):
            html = html.encode('utf-8')
        response = HtmlResponse(url, encoding='utf-8', body=html)
        items = dict()
        if not is_detail:
            if column_name in ['article', 'article_urls']:
                items[column_name] = list()
                if not (xpath and regex):
                    links = NewsLinkExtractor().calc_top_link(url=url, node=html)[0]
                    for link in links.values():
                        items[column_name].append(str(link))
                if xpath:
                    links = LxmlLinkExtractor(restrict_xpaths=xpath).extract_links(response)
                    if links:
                        items[column_name] = []
                    for link in links:
                        items[column_name].append(link.url)
                if regex:
                    links = LxmlLinkExtractor(allow=[regex]).extract_links(response)
                    if links:
                        items[column_name] = []
                    for link in links:
                        items[column_name].append(link.url)
            elif column_name.startswith('list_'):
                if xpath:
                    html = Selector(response)
                    if html.xpath(xpath).extract():
                        try:
                            extract_result = html.xpath(xpath).extract()
                            items[column_name] = extract_result
                        except IndexError as E:
                            logger.error(
                                'extract error %s %s %s %s' % (column_name, xpath, url, traceback.format_exc()),
                                extra=data)
                elif regex:
                    try:
                        items[column_name] = re.findall(regex, response.text)
                    except IndexError as E:
                        logger.error('extract error %s %s %s %s' % (column_name, xpath, url, traceback.format_exc()),
                                     extra=data)
        else:
            if xpath:
                html = Selector(response)
                if html.xpath(xpath).extract():
                    try:
                        extract_result = html.xpath(xpath).extract()
                        if column_name == "content_html":
                            items[column_name] = "\n".join(extract_result)
                            items['content'] = Clean.clean_blank_line(items[column_name])
                        elif column_name == "attachment":
                            items[column_name] = Toolkit.extract_attachment(response.text, url, xpath)
                        else:
                            items[column_name] = extract_result[0]
                    except IndexError as E:
                        logger.error(
                            'extract error %s %s %s %s' % (column_name, xpath, url, traceback.format_exc()),
                            extra=data)
            elif regex:
                try:
                    items[column_name] = re.findall(regex, response.text)[0]
                except IndexError as E:
                    logger.error('extract error %s %s %s %s' % (column_name, xpath, url, traceback.format_exc()),
                                 extra=data)
        return items


class NewExtractor(object):
    @staticmethod
    def web_collector(url, html, extract_server):
        url = url
        if isinstance(html, str):
            html = html.encode('utf-8')
        html = {'html': html}
        extract_response = requests.post(extract_server, data=html,
                                         headers={'Content-Type': 'application/x-www-form-urlencoded',
                                                  'accept-language': 'en'}, verify=False, timeout=50)
        EXTRACT_COUNTER.inc()  # point
        items = dict()
        scrapy_response = HtmlResponse(url, encoding='utf-8', body=extract_response.content)
        jsr = json.loads(scrapy_response.text)
        tc = NewsExtractor()
        tc_result = tc.calc_main_block(url=url, node=html.get('html'))
        if jsr.get('title'):
            items['title'] = jsr.get('title')
        else:
            items['title'] = tc_result[0]
        if tc_result[1]:
            items['pub_time'] = tc_result[1]
        else:
            items['pub_time'] = jsr.get('time')[0:10]
        if tc_result[2]:
            items['content_html'] = Clean.clean_blank_line(tc_result[2])
        else:
            items['content_html'] = Clean.clean_blank_line(jsr.get('html'))
        items['content'] = Clean.clean_html_tag(items['content_html'])
        return items

    @staticmethod
    def extract(url, column_name, xpath, regex, html, is_link, data, get_all):
        if isinstance(html, str):
            html = html.encode('utf-8')
        response = HtmlResponse(url, encoding='utf-8', body=html)
        items = dict()
        # if not (xpath and regex):
        #     tc = NewsExtractor()
        #     tc_result = tc.calc_main_block(url=url, node=html)
        #     items['title'] = tc_result[0]
        #     if tc_result[1]:
        #         items['pub_time'] = tc_result[1]
        #     if tc_result[2]:
        #         items['content_html'] = Clean.clean_blank_line(tc_result[2])
        #     items['content'] = Clean.clean_html_tag(items['content_html'])
        if is_link:
            items[column_name] = list()
            if not (xpath and regex):
                links = NewsLinkExtractor().calc_top_link(url=url, node=html)[0]
                if isinstance(links, dict):
                    for link in links.values():
                        items[column_name].append(str(link))
            if xpath:
                # xpath = ''.join(xpath.split('/@href')[0:-1])
                links = LxmlLinkExtractor(restrict_xpaths=xpath).extract_links(response)
                if links:
                    items[column_name] = []
                for link in links:
                    items[column_name].append(link.url)
            if regex:
                links = LxmlLinkExtractor(allow=[regex]).extract_links(response)
                for link in links:
                    items[column_name].append(link.url)
        else:
            if xpath:
                html = Selector(response)
                if html.xpath(xpath).extract():
                    try:
                        extract_result = html.xpath(xpath).extract()
                        if column_name == "content_html":
                            items[column_name] = "\n".join(extract_result)
                            items['content'] = Clean.clean_blank_line(items[column_name])
                        elif column_name == "attachment":
                            items[column_name] = Toolkit.extract_attachment(response.text, url, xpath)
                        else:
                            if not get_all:
                                items[column_name] = extract_result[0]
                            else:
                                items[column_name] = extract_result
                    except IndexError as E:
                        logger.error(
                            'extract error %s %s %s %s' % (column_name, xpath, url, traceback.format_exc()),
                            extra=data)
            elif regex:
                try:
                    if not get_all:
                        items[column_name] = re.findall(regex, response.text)[0]
                    else:
                        items[column_name] = re.findall(regex, response.text)
                except IndexError as E:
                    logger.error('extract error %s %s %s %s' % (column_name, xpath, url, traceback.format_exc()),
                                 extra=data)

        return items
